from itertools import product
import json
import os
from openslide import OpenSlide
from tqdm import tqdm
import zarr

def tiff_to_zarr(tiff_path: str, zarr_path: str, meta_path: str, tile_size: int = 4096, chunk_size: int = 4096) -> None:
    assert isinstance(tiff_path, str), 'tiff_path must be a string'
    assert os.path.exists(tiff_path), 'tiff_path does not exist'
    assert isinstance(zarr_path, str), 'zarr_path must be a string'
    assert not os.path.exists(zarr_path), 'zarr_path already exists'
    assert isinstance(meta_path, str), 'meta_path must be a string'
    assert not os.path.exists(meta_path), 'meta_path already exists'
    assert isinstance(tile_size, int), 'tile_size must be an integer'
    assert tile_size > 0, 'tile_size must be positive'
    assert isinstance(chunk_size, int), 'chunk_size must be an integer'
    assert chunk_size > 0, 'chunk_size must be positive'
    with OpenSlide(tiff_path) as slide:
        slide_dims = slide.dimensions
        out_shape = [slide_dims[1], slide_dims[0], 3]
        out_img = zarr.zeros(out_shape, dtype='uint8', chunks=(chunk_size, chunk_size, 3), store = zarr.DirectoryStore(zarr_path))
        y_mins = range(0, out_shape[0], tile_size)
        x_mins = range(0, out_shape[1], tile_size)
        num_iter = len(y_mins) * len(x_mins)
        for y_min, x_min in tqdm(product(y_mins, x_mins), total=num_iter):
            y_max = min(y_min + tile_size, out_shape[0])
            x_max = min(x_min + tile_size, out_shape[1])
            tile = slide.read_region((x_min, y_min), 0, (x_max - x_min, y_max - y_min)).convert('RGB')
            out_img[y_min:y_max, x_min:x_max] = tile
        slide_metadata = {
            'name': os.path.basename(tiff_path),
            'format': str(slide.detect_format(tiff_path)),
        }
        if hasattr(slide, 'properties') and 'openslide.mpp-x' in slide.properties and 'openslide.mpp-y' in slide.properties:
            slide_metadata['mpp-x'] = float(slide.properties['openslide.mpp-x'])
            slide_metadata['mpp-y'] = float(slide.properties['openslide.mpp-y'])
            if slide_metadata['mpp-x'] == slide_metadata['mpp-y']:
                slide_metadata['mpp'] = slide_metadata['mpp-x']
        with open(meta_path, 'w') as f:
            json.dump(slide_metadata, f)


"""
Profile for connecting to Eddie, the University of Edinburgh's HPC cluster.

Profile contact: Hugh Warden - h.b.warden@sms.ed.ac.uk
"""

eddie_profile = {
    "cluster_type": "SGE",
    "memory": '8 G',
    "cores": 1,
    "resource_spec": "h_vmem=8G",
    "worker_extra_args": ["--lifetime", "25m", "--lifetime-stagger", "4m"]
}


from .._process import Process
from ...image._image import Image
from ...image._ubyte_image import UByteImage
from dask_image.ndfilters import gaussian_filter

class GaussianBlur(Process):
    def __init__(self, sigma: float, metric: str = 'micron', truncate: float = 3.0) -> None:
        assert isinstance(sigma, (int, float)), 'sigma must be a number'
        assert sigma > 0, 'sigma must be positive'
        assert isinstance(metric, str), 'metric must be a string'
        assert metric in ['micron', 'pixel'], 'metric must be either "micron" or "pixel"'
        assert isinstance(truncate, (int, float)), 'truncate must be a number'
        assert truncate > 0, 'truncate must be positive'
        self.sigma = sigma
        self.metric = metric
        self.truncate = truncate
        super().__init__(name=f'Gaussian Blur with a sigma of {sigma} {metric}s')

    def run(self, image: Image) -> Image:
        assert isinstance(image, Image), 'image must be an Image'
        sigma = image.pixel_from_micron(self.sigma) if self.metric == 'micron' else self.sigma
        truncate = self.truncate
        img = image.image
        if isinstance(image, UByteImage):
            img = img.astype('float32')
        img = gaussian_filter(img, sigma=(sigma, sigma, 0), mode='mirror', truncate=truncate)
        if isinstance(image, UByteImage):
            img = img.astype('uint8')
            return UByteImage(img, image.info, image.channel_names)
        else:
            raise NotImplementedError('GaussianBlur only supports UByteImage objects at this time')



#   -------------------------------------------------------------
#   Copyright (c) Microsoft Corporation. All rights reserved.
#   Licensed under the MIT License. See LICENSE in project root for information.
#   -------------------------------------------------------------
"""Python Package Template"""
from __future__ import annotations

__version__ = "0.0.2-rc26-post1"



from ..image._ubyte_image import UByteImage
import dask.array as da
import json
import os
from typing import List

def read_zarr_img(zarr_path: str, meta_path: str, channel_names: List[str] = ["Red", "Green", "Blue"]) -> UByteImage:
    assert isinstance(zarr_path, str)
    assert zarr_path.endswith('.zarr')
    assert os.path.exists(zarr_path)
    assert isinstance(meta_path, str)
    assert meta_path.endswith('.json')
    assert os.path.exists(meta_path)
    img = da.from_zarr(zarr_path)
    with open(meta_path, 'r') as f:
        meta = json.load(f)
    if str(img.dtype) == "uint8":
        return UByteImage(img, meta, channel_names)
    else:
        raise NotImplementedError(f"Images of type {str(img.dtype)} are not supported yet.")


from ._distribution_filters import MaximumFilter, MedianFilter, MinimumFilter, PercentileFilter
from ._gaussian_blur import GaussianBlur

__all__ = [
    'GaussianBlur',
    'MaximumFilter',
    'MedianFilter',
    'MinimumFilter',
    'PercentileFilter'
]


from abc import ABC, abstractmethod
from dask.array import Array
import matplotlib.pyplot as plt
import numpy as np
from numpy import ndarray
import os
from typing import List

class Image(ABC):
    def __init__(self, image: Array, info: dict, channel_names: List[str]) -> None:
        assert isinstance(image, Array), "image must be a dask array"
        assert image.ndim in [2, 3], "image must be 2D or 3D"
        assert isinstance(info, dict), "info must be a dictionary"
        if isinstance(channel_names, str):
            channel_names = [channel_names]
        assert isinstance(channel_names, list), "channel_names must be a list"
        assert all(isinstance(name, str) for name in channel_names), "all channel names must be strings"
        if image.ndim == 3:
            assert len(channel_names) == image.shape[2], "number of channel names must match number of channels in image"
        else:
            assert len(channel_names) == 1, "number of channel names must match number of channels in image"
        self.image = image
        self.info = info
        self.ndim = image.ndim
        self.shape = image.shape
        self.dtype = str(image.dtype)
        self.name = info.get('name', None)
        if 'mpp' in self.info:
            self.mpp = self.info['mpp']
        elif 'mpp-x' in self.info and 'mpp-y' in self.info:
            if self.info['mpp-x'] == self.info['mpp-y']:
                self.mpp = self.info['mpp-x']
        else:
            self.mpp = None
        self.channel_names = channel_names

    def persist(self) -> None:
        return self.image.persist()

    def compute(self) ->  None:
        return self.image.compute()

    def rechunk(self, chunks: tuple) -> None:
        assert isinstance(chunks, tuple), "chunks must be a tuple"
        assert all(isinstance(chunk, int) for chunk in chunks), "all elements of chunks must be integers"
        assert len(chunks) == self.ndim, "length of chunks must match number of dimensions in image"
        assert all(chunk > 0 for chunk in chunks), "all elements of chunks must be positive"
        if self.ndim == 3:
            assert chunks[2] == len(self.channel_names), "third element of chunks must match number of channels in image"
        self.image = self.image.rechunk(chunks)

    def pixel_from_micron(self, micron: float) -> float:
        assert isinstance(micron, (int, float)), "micron must be a number"
        assert micron > 0, "micron must be positive"
        if not self.mpp or not isinstance(self.mpp, (int, float)):
            raise ValueError("Microns per pixel (mpp) not available")
        return micron / self.mpp

    def read_region(self, y_min: int, x_min: int, y_len: int, x_len: int, pad: bool = False) -> ndarray:
        assert isinstance(y_min, int), "y_min must be an integer"
        assert isinstance(x_min, int), "x_min must be an integer"
        assert isinstance(y_len, int), "y_len must be an integer"
        assert isinstance(x_len, int), "x_len must be an integer"
        assert isinstance(pad, bool), "pad must be a boolean"
        assert y_min >= 0, "y_min must be non-negative"
        assert x_min >= 0, "x_min must be non-negative"
        assert y_len > 0, "y_len must be positive"
        assert x_len > 0, "x_len must be positive"
        assert y_min + y_len <= self.shape[0] or pad, "y_min + y_len must be less than or equal to the height of the image if image is not padded"
        assert x_min + x_len <= self.shape[1] or pad, "x_min + x_len must be less than or equal to the width of the image if image is not padded"
        y_max = min(y_min+y_len, self.shape[0])
        x_max = min(x_min+x_len, self.shape[1])
        region = self.image[y_min:y_max, x_min:x_max].compute()
        if pad and region.shape[:2] != (y_len, x_len):
            pad_y = y_len - region.shape[0]
            pad_x = x_len - region.shape[1]
            if self.ndim == 2:
                region = np.pad(region, ((0, pad_y), (0, pad_x)), mode='constant', constant_values=0)
            elif self.ndim == 3:
                region = np.pad(region, ((0, pad_y), (0, pad_x), (0, 0)), mode='constant', constant_values=0)
        return region

    def show_region(self, y_min: int, x_min: int, y_len: int, x_len:int, pad: bool = False) -> None:
        region = self.read_region(y_min, x_min, y_len, x_len, pad)
        plt.imshow(region, cmap='gray' if self.ndim == 2 or len(self.channel_names) == 1 else None)
        plt.axis('off')
        plt.show()

    @abstractmethod
    def get_thumb(self):
        pass

    @abstractmethod
    def show_thumb(self) ->  None:
        pass

    def _repr_html_(self) -> str:
        colour_map = {
            "Red": "#FF0000",
            "Green": "#00FF00",
            "Blue": "#0000FF",
            "Hematoxylin": "#2A2670",
            "Eosin": "#E63DB7",
            "DAB": "#813723"
        }
        channel_names = []
        for colour in self.channel_names:
            if colour in colour_map:
                channel_names.append(f'<span style="color: {colour_map[colour]};">{colour}</span>')
            else:
                channel_names.append(colour)
        total_width = 400
        icon_url = 'https://github.com/SCEMA-WSI/scematk/blob/main/src/scematk/image/_icons/temp_icon.png'
        html = f'''
        <div style="width: {total_width}px; background-color: #202020; padding: 20px; border-radius: 5px; box-shadow: 0 2px 4px rgba(0, 0, 0, 0.1);">
            <h1>SCEMATK Image Object</h1>
            <div style="display: flex; justify-content: space-between;">
                <div style="flex: 2;">
                    <p>Name: {self.name if self.name else "SCEMATK Image"}</p>
                    <p>Format: {self.info.get('format', 'Unknown').title()}</p>
                    <p>Channels: {", ".join(channel_names)}</p>
                    <p>Dimensions: {" x ".join([f"{i:,}" for i in self.shape])}</p>
                    <p>Chunks: {" x ".join([f"{i:,}" for i in self.image.chunksize[:2]])}
                </div>
                <div style="flex: 1;">
                    <img src="{icon_url}" style="max-width: 100%; height: auto; margin-top: 10px;">
                </div>
            </div>
        </div>
        '''
        return html
